import json
import math
from os import path

import numpy as np
import pytest

from truthdiscovery.algorithm import (
    AverageLog,
    BaseIterativeAlgorithm,
    Investment,
    MajorityVoting,
    PooledInvestment,
    PriorBelief,
    Sums,
    TruthFinder
)
from truthdiscovery.exceptions import EmptyDatasetError
from truthdiscovery.input import Dataset, MatrixDataset
from truthdiscovery.utils import (
    ConvergenceIterator,
    DistanceMeasures,
    FixedIterator
)


class BaseTest:
    @pytest.fixture
    def triples(self):
        return [
            ("s1", "x", "one"),
            ("s1", "y", "nine"),
            ("s1", "z", "seven"),

            ("s2", "x", "one"),
            ("s2", "y", "eight"),

            ("s3", "z", "seven"),
        ]

    @pytest.fixture
    def data(self, triples):
        return Dataset(triples)


class TestBase(BaseTest):
    def test_empty_dataset(self):
        data = Dataset([])
        non_it = MajorityVoting()
        it = Sums()
        for alg in [non_it, it]:
            with pytest.raises(EmptyDatasetError) as excinfo:
                alg.run(data)
        err_msg = "Cannot run algorithm on empty dataset"
        assert str(excinfo.value) == err_msg

        # Test with run_iter also
        with pytest.raises(EmptyDatasetError) as excinfo2:
            _l = list(it.run_iter(data))
        assert str(excinfo2.value) == err_msg

    def test_get_parameter_names(self):
        assert MajorityVoting.get_parameter_names() == set([])
        assert PooledInvestment.get_parameter_names() == {
            "priors", "iterator", "g"
        }
        assert TruthFinder.get_parameter_names() == {
            "priors", "iterator", "influence_param", "dampening_factor",
            "initial_trust"
        }


class TestVoting(BaseTest):
    def test_basic(self, data):
        voting = MajorityVoting()
        results = voting.run(data)
        assert results.trust == {"s1": 1, "s2": 1, "s3": 1}
        assert results.belief == {
            "x": {"one": 1},
            "y": {"nine": 0.5, "eight": 0.5},
            "z": {"seven": 1}
        }


class TestBaseIterative(BaseTest):
    def test_run_fail(self, data):
        alg = BaseIterativeAlgorithm()
        with pytest.raises(NotImplementedError):
            alg.run(data)

    def test_fixed_priors(self, data):
        alg = BaseIterativeAlgorithm(priors=PriorBelief.FIXED)
        got = alg.get_prior_beliefs(data)
        # Claims are:
        # 0: x=1
        # 1: y=9
        # 2: z=7
        # 3: y=8
        expected = [0.5, 0.5, 0.5, 0.5]
        assert np.array_equal(got, expected)

    def test_voted_priors(self, data, triples):
        alg = BaseIterativeAlgorithm(priors=PriorBelief.VOTED)
        got = alg.get_prior_beliefs(data)
        expected = [1, 0.5, 1, 0.5]
        assert np.array_equal(got, expected)

        # Add a new value so that priors are different to voted...
        data2 = Dataset(triples + [("s3", "y", "eight")])
        got2 = alg.get_prior_beliefs(data2)
        expected2 = [1, 1 / 3, 1, 2 / 3]
        assert np.array_equal(got2, expected2)

    def test_uniform_priors(self, data, triples):
        alg = BaseIterativeAlgorithm(priors=PriorBelief.UNIFORM)
        got = alg.get_prior_beliefs(data)
        expected = [1, 0.5, 1, 0.5]
        assert np.array_equal(got, expected)

        # Same secondary test as for voted, but beliefs should not change here
        data2 = Dataset(triples + [("s3", "y", "eight")])
        got2 = alg.get_prior_beliefs(data2)
        expected2 = [1, 0.5, 1, 0.5]
        assert np.array_equal(got2, expected2)

    def test_invalid_priors(self, data):
        alg = BaseIterativeAlgorithm(priors="hello")
        with pytest.raises(ValueError):
            alg.get_prior_beliefs(data)


class TestSums(BaseTest):
    def test_basic(self, data):
        """
        Perform Sums on a small graph. The expected results were obtained by
        finding eigenvectors of suitable matrices (using numpy "by hand"), as
        per Kleinberg paper for Hubs and Authorities
        """
        sums = Sums(iterator=ConvergenceIterator(DistanceMeasures.L1, 0.00001))
        results = sums.run(data)
        assert np.isclose(results.trust["s1"], 1)
        assert np.isclose(results.trust["s2"], 0.53208889)
        assert np.isclose(results.trust["s3"], 0.34729636)

        assert set(results.belief["x"].keys()) == {"one"}
        assert np.isclose(results.belief["x"]["one"], 1)

        assert set(results.belief["y"].keys()) == {"eight", "nine"}
        assert np.isclose(results.belief["y"]["nine"], 0.65270364)
        assert np.isclose(results.belief["y"]["eight"], 0.34729636)

        assert set(results.belief["z"].keys()) == {"seven"}
        assert np.isclose(results.belief["z"]["seven"], 0.87938524)


class TestAverageLog(BaseTest):
    def test_basic(self, data):
        num_iterations = 20
        avlog = AverageLog(iterator=FixedIterator(num_iterations))
        results = avlog.run(data)

        t = [0, 0, 0]
        b = [0.5, 0.5, 0.5, 0.5]

        for _ in range(num_iterations):
            t[0] = math.log(3) * (b[0] + b[1] + b[3]) / 3
            t[1] = math.log(2) * (b[0] + b[2]) / 2
            t[2] = 0

            b[0] = t[0] + t[1]
            b[1] = t[0]
            b[2] = t[1]
            b[3] = t[0] + t[2]

            max_t = max(t)
            max_b = max(b)
            for i, val in enumerate(t):
                t[i] = val / max_t
            for j, val in enumerate(b):
                b[j] = val / max_b

        assert np.isclose(results.trust["s1"], t[0])
        assert np.isclose(results.trust["s2"], t[1])
        assert np.isclose(results.trust["s3"], t[2])

        assert set(results.belief["x"].keys()) == {"one"}
        assert set(results.belief["y"].keys()) == {"eight", "nine"}
        assert set(results.belief["z"].keys()) == {"seven"}

        assert np.isclose(results.belief["x"]["one"], b[0])
        assert np.isclose(results.belief["y"]["nine"], b[1])
        assert np.isclose(results.belief["y"]["eight"], b[2])
        assert np.isclose(results.belief["z"]["seven"], b[3])


class TestInvestment(BaseTest):
    def test_basic(self):
        data = Dataset([
            ("s1", "x", "one"),
            ("s2", "x", "zero"),
            ("s3", "x", "one"),

            ("s1", "y", "zero"),
            ("s3", "y", "one"),
            ("s4", "y", "one"),

            ("s2", "z", "zero"),
            ("s3", "z", "one")
        ])
        # TODO: find example that converges nicely: this one results in all
        # sources bar one having trust 0, which causes division by zero error
        num_iterations = 20
        g = 1.4
        inv = Investment(iterator=FixedIterator(num_iterations), g=g)
        results = inv.run(data)

        b = [
            2 / 3,  # x = 1
            1 / 3,  # x = 0
            2 / 3,  # y = 1
            1 / 3,  # y = 0
            1 / 2,  # z = 1
            1 / 2   # z = 0
        ]
        t = [1, 1, 1, 1]
        old_t = t[:]

        for _ in range(num_iterations):
            t[0] = (old_t[0] / 2) * (
                b[0] / ((old_t[0] / 2) + (old_t[2] / 3)) +
                b[3] / (old_t[0] / 2)
            )
            t[1] = (old_t[1] / 2) * (
                b[1] / (old_t[1] / 2) +
                b[5] / (old_t[1] / 2)
            )
            t[2] = (old_t[2] / 3) * (
                b[0] / ((old_t[0] / 2) + (old_t[2] / 3)) +
                b[2] / ((old_t[2] / 3) + old_t[3]) +
                b[4] / (old_t[2] / 3)
            )
            t[3] = old_t[3] * (
                b[2] / ((old_t[2] / 3) + old_t[3])
            )

            b[0] = ((t[0] / 2) + (t[2] / 3)) ** g
            b[1] = (t[1] / 2) ** g
            b[2] = ((t[2] / 3) + t[3]) ** g
            b[3] = (t[0] / 2) ** g
            b[4] = (t[2] / 3) ** g
            b[5] = (t[1] / 2) ** g

            max_t = max(t)
            for i, val in enumerate(t):
                t[i] = val / max_t
            old_t = t[:]

            max_b = max(b)
            for j, val in enumerate(b):
                b[j] = val / max_b

        assert np.isclose(results.trust["s1"], t[0])
        assert np.isclose(results.trust["s2"], t[1])
        assert np.isclose(results.trust["s3"], t[2])

        assert set(results.belief["x"].keys()) == {"zero", "one"}
        assert set(results.belief["y"].keys()) == {"zero", "one"}
        assert set(results.belief["z"].keys()) == {"zero", "one"}

        assert np.isclose(results.belief["x"]["one"], b[0])
        assert np.isclose(results.belief["x"]["zero"], b[1])
        assert np.isclose(results.belief["y"]["one"], b[2])
        assert np.isclose(results.belief["y"]["zero"], b[3])
        assert np.isclose(results.belief["z"]["one"], b[4])
        assert np.isclose(results.belief["z"]["zero"], b[5])

    def test_converge_to_zero(self):
        """"
        Run Investment till convergence with input that is known to cause
        problems as trust drains to zero, which causes division by zero.
        """
        data = Dataset([
            ("s1", "x", "one"),
            ("s2", "x", "zero"),
            ("s3", "x", "one"),

            ("s1", "y", "zero"),
            ("s3", "y", "one"),
            ("s4", "y", "one"),

            ("s2", "z", "zero"),
            ("s3", "z", "one")
        ])
        it = ConvergenceIterator(DistanceMeasures.L2, 0.1e-100)
        res = Investment(iterator=it).run(data)
        assert res.iterations == 41


class TestPooledInvestment(BaseTest):
    def test_basic(self):
        # Same data as Investment above
        data = Dataset([
            ("s1", "x", "one"), ("s2", "x", "zero"), ("s3", "x", "one"),
            ("s1", "y", "zero"), ("s3", "y", "one"), ("s4", "y", "one"),
            ("s2", "z", "zero"), ("s3", "z", "one")
        ])

        num_iterations = 5
        g = 1.4
        pooled_inv = PooledInvestment(
            iterator=FixedIterator(num_iterations), g=g
        )
        results = pooled_inv.run(data)

        b = [0.5] * 6
        t = [1, 1, 1, 1]
        old_t = t[:]

        for _ in range(num_iterations):
            # Note: trust update is the same as for investment
            t[0] = (old_t[0] / 2) * (
                b[0] / ((old_t[0] / 2) + (old_t[2] / 3)) +
                b[3] / (old_t[0] / 2)
            )
            t[1] = (old_t[1] / 2) * (
                b[1] / (old_t[1] / 2) +
                b[5] / (old_t[1] / 2)
            )
            t[2] = (old_t[2] / 3) * (
                b[0] / ((old_t[0] / 2) + (old_t[2] / 3)) +
                b[2] / ((old_t[2] / 3) + old_t[3]) +
                b[4] / (old_t[2] / 3)
            )
            t[3] = old_t[3] * (
                b[2] / ((old_t[2] / 3) + old_t[3])
            )

            h = [
                (t[0] / 2) + (t[2] / 3),
                t[1] / 2,
                (t[2] / 3) + t[3],
                t[0] / 2,
                t[2] / 3,
                t[1] / 2,
            ]
            g_h = [x ** g for x in h]

            b[0] = h[0] * g_h[0] / (g_h[0] + g_h[1])
            b[1] = h[1] * g_h[1] / (g_h[0] + g_h[1])
            b[2] = h[2] * g_h[2] / (g_h[2] + g_h[3])
            b[3] = h[3] * g_h[3] / (g_h[2] + g_h[3])
            b[4] = h[4] * g_h[4] / (g_h[4] + g_h[5])
            b[5] = h[5] * g_h[5] / (g_h[4] + g_h[5])

            max_t = max(t)
            for i, val in enumerate(t):
                t[i] = val / max_t
            old_t = t[:]

            max_b = max(b)
            for j, val in enumerate(b):
                b[j] = val / max_b

        assert np.isclose(results.trust["s1"], t[0])
        assert np.isclose(results.trust["s2"], t[1])
        assert np.isclose(results.trust["s3"], t[2])

        assert set(results.belief["x"].keys()) == {"zero", "one"}
        assert set(results.belief["y"].keys()) == {"zero", "one"}
        assert set(results.belief["z"].keys()) == {"zero", "one"}

        assert np.isclose(results.belief["x"]["one"], b[0])
        assert np.isclose(results.belief["x"]["zero"], b[1])
        assert np.isclose(results.belief["y"]["one"], b[2])
        assert np.isclose(results.belief["y"]["zero"], b[3])
        assert np.isclose(results.belief["z"]["one"], b[4])
        assert np.isclose(results.belief["z"]["zero"], b[5])


class TestTruthFinder(BaseTest):
    """
    Test the TruthFinder algorithm
    """
    def test_basic(self, triples):
        def imp_func(var, val1, val2):
            m = {"eight": 8, "nine": 9}
            # Make sure implication is asymmetric
            diff = m[val1] - m[val2]
            if diff > 0:
                return np.exp(-0.5 * diff * diff)
            return 0.4

        imp_data = Dataset(triples, implication_function=imp_func)
        n = 50
        t_0 = 0.4
        gamma = 0.5
        ro = 0.25

        trust = [t_0, t_0, t_0]
        belief = [0] * 4

        for _ in range(n):
            tau = [
                -np.log(1 - trust[0]),
                -np.log(1 - trust[1]),
                -np.log(1 - trust[2])
            ]
            sigma = [
                tau[0] + tau[1],
                tau[0],
                tau[1],
                tau[0] + tau[2]
            ]
            sigma_star = [
                sigma[0],
                sigma[1] + ro * sigma[2] * 0.4,
                sigma[2] + ro * sigma[1] * np.exp(-0.5),
                sigma[3]
            ]
            belief = [
                1 / (1 + np.exp(-gamma * sigma_star[0])),
                1 / (1 + np.exp(-gamma * sigma_star[1])),
                1 / (1 + np.exp(-gamma * sigma_star[2])),
                1 / (1 + np.exp(-gamma * sigma_star[3]))
            ]
            trust = [
                (belief[0] + belief[1] + belief[3]) / 3,
                (belief[0] + belief[2]) / 2,
                belief[3]
            ]

        truthfinder = TruthFinder(
            dampening_factor=gamma, influence_param=ro, initial_trust=t_0,
            iterator=FixedIterator(n)
        )
        results = truthfinder.run(imp_data)

        assert np.isclose(results.trust["s1"], trust[0])
        assert np.isclose(results.trust["s2"], trust[1])
        assert np.isclose(results.trust["s3"], trust[2])

        assert set(results.belief["x"].keys()) == {"one"}
        assert set(results.belief["y"].keys()) == {"eight", "nine"}
        assert set(results.belief["z"].keys()) == {"seven"}

        assert np.isclose(results.belief["x"]["one"], belief[0])
        assert np.isclose(results.belief["y"]["nine"], belief[1])
        assert np.isclose(results.belief["y"]["eight"], belief[2])
        assert np.isclose(results.belief["z"]["seven"], belief[3])

    def test_no_implications(self, data):
        """
        Perform the same run as above, but do not bother with implications
        between claims. This is to check that implications are ignored if no
        implication function is given
        """
        n = 50
        t_0 = 0.4
        gamma = 0.5

        trust = [t_0, t_0, t_0]
        belief = [0] * 4

        for _ in range(n):
            tau = [
                -np.log(1 - trust[0]),
                -np.log(1 - trust[1]),
                -np.log(1 - trust[2])
            ]
            sigma = [
                tau[0] + tau[1],
                tau[0],
                tau[1],
                tau[0] + tau[2]
            ]
            belief = [
                1 / (1 + np.exp(-gamma * sigma[0])),
                1 / (1 + np.exp(-gamma * sigma[1])),
                1 / (1 + np.exp(-gamma * sigma[2])),
                1 / (1 + np.exp(-gamma * sigma[3]))
            ]
            trust = [
                (belief[0] + belief[1] + belief[3]) / 3,
                (belief[0] + belief[2]) / 2,
                belief[3]
            ]

        truthfinder = TruthFinder(
            dampening_factor=gamma, initial_trust=t_0,
            iterator=FixedIterator(n)
        )
        results = truthfinder.run(data)

        assert np.isclose(results.trust["s1"], trust[0])
        assert np.isclose(results.trust["s2"], trust[1])
        assert np.isclose(results.trust["s3"], trust[2])

        assert set(results.belief["x"].keys()) == {"one"}
        assert set(results.belief["y"].keys()) == {"eight", "nine"}
        assert set(results.belief["z"].keys()) == {"seven"}

        assert np.isclose(results.belief["x"]["one"], belief[0])
        assert np.isclose(results.belief["y"]["nine"], belief[1])
        assert np.isclose(results.belief["y"]["eight"], belief[2])
        assert np.isclose(results.belief["z"]["seven"], belief[3])

    def test_trust_invalid(self):
        """
        In theory trust scores cannot be 1 for any source. In practise scores
        get so close to 1 that they are rounded to 1, which causes problems
        when we do log(1 - trust).

        This test checks that iteration stops in this case
        """
        data = MatrixDataset(np.array([
            [1, 2, 3],
            [1, 2, 3],
            [1, 2, 3],
            [1, 2, 3],
            [1, 2, 3]
        ]))
        it = FixedIterator(100)
        alg = TruthFinder(iterator=it)
        res = alg.run(data)
        # Iteration should stop after only 7 iterations, instead of 100
        assert it.it_count == 7
        assert res.iterations == 7


class TestOnLargeData:
    """
    The following are regression tests, to check that the output of each
    algorithm on a (synthetic) dataset remains the same over time
    """
    @pytest.fixture
    def data(self):
        data_path = self.get_filepath("data.csv")
        with open(data_path) as csv_file:
            return MatrixDataset.from_csv(csv_file)

    def get_filepath(self, name):
        here = path.abspath(path.dirname(__file__))
        return path.join(here, "regression", name)

    def check_results(self, alg, data, exp_results_filename):
        """
        Helper method to check results of an algorithm

        :param alg: algorithm instance
        :param data: Dataset instance
        :param exp_results_filename: filename of JSON file containing expected
                                     results (name should be relative to
                                     'truthdiscovery/test/regression' in the
                                     repo)
        """
        res = alg.run(data)
        with open(self.get_filepath(exp_results_filename)) as res_file:
            exp_res = json.load(res_file)

        for source, trust_val in exp_res["trust"].items():
            source = int(source)
            err_msg = "Incorrect trust for {}, source {}".format(
                exp_results_filename, source
            )
            assert np.isclose(res.trust[source], trust_val), err_msg

        for var, beliefs in exp_res["belief"].items():
            var = int(var)
            for val, belief_val in beliefs.items():
                val = float(val)
                err_msg = "Incorrect belief for {}, var {}, value {}".format(
                    exp_results_filename, var, val
                )
                assert np.isclose(res.belief[var][val], belief_val), err_msg

    def test_sums(self, data):
        sums = Sums(iterator=FixedIterator(20))
        self.check_results(sums, data, "sums_results.json")

    def test_average_log(self, data):
        average_log = AverageLog(iterator=FixedIterator(20))
        self.check_results(average_log, data, "average_log_results.json")

    def test_investment(self, data):
        investment = Investment(iterator=FixedIterator(20))
        self.check_results(investment, data, "investment_results.json")

    def test_pooled_investment(self, data):
        pooled_investment = PooledInvestment(iterator=FixedIterator(10))
        self.check_results(
            pooled_investment, data, "pooled_investment_results.json"
        )

    def test_truthfinder(self, data):
        it = ConvergenceIterator(DistanceMeasures.COSINE, 0.001)
        truthfinder = TruthFinder(iterator=it)

        def imp(var, val1, val2):
            diff = val1 - val2
            return np.exp(-0.5 * diff ** 2)

        data = MatrixDataset(data.sv, implication_function=imp)
        self.check_results(truthfinder, data, "truthfinder_results.json")

    def test_voting(self, data):
        voting = MajorityVoting()
        self.check_results(voting, data, "voting_results.json")


class TestIteratorsForAlgorithms:
    def test_default_iterator_types(self):
        test_data = {
            FixedIterator: (Sums, AverageLog, Investment, PooledInvestment),
            ConvergenceIterator: (TruthFinder,)
        }
        for it_cls, alg_classes in test_data.items():
            for alg_cls in alg_classes:
                obj = alg_cls()
                err_msg = ("Wrong iterator type for {}: {} instead of {}"
                           .format(alg_cls.__name__,
                                   obj.iterator.__class__.__name__,
                                   it_cls.__name__))
                assert isinstance(obj.iterator, it_cls), err_msg


class TestLoggingAlgorithm(BaseTest):
    @pytest.fixture
    def alg_classes(self):
        return [AverageLog, Investment, PooledInvestment, Sums, TruthFinder]

    @pytest.fixture
    def algs(self, alg_classes):
        return [cls() for cls in alg_classes]

    def test_final_result(self, algs, data):
        for alg in algs:
            # Get the final result normally
            final_res = alg.run(data)

            # Iterate through partial results, and check the final result is as
            # expected
            last_res = None
            for r in alg.run_iter(data):
                last_res = r

            assert last_res.trust == final_res.trust
            assert last_res.belief == final_res.belief
            assert last_res.iterations == final_res.iterations

    def test_no_logging(self, algs, data):
        for alg in algs:
            _res = alg.run(data)
            assert alg.results_log is None

    def test_partial_results(self, alg_classes, data):
        it = FixedIterator(3)
        for cls in alg_classes:
            alg = cls(iterator=it)

            log = list(alg.run_iter(data))
            assert len(log) == 4
            initial, first, second, third = log

            assert initial.iterations == 0
            assert first.iterations == 1
            assert second.iterations == 2
            assert third.iterations == 3

            # Time should be non-decreasing
            assert (
                initial.time_taken <= first.time_taken <= second.time_taken
                <= third.time_taken
            )

    def test_sums_detailed(self, data):
        it = FixedIterator(3)
        alg = Sums(iterator=it, priors=PriorBelief.FIXED)
        initial, first, second, third = alg.run_iter(data)

        assert initial.belief == {
            "x": {"one": 0.5},
            "y": {"nine": 0.5, "eight": 0.5},
            "z": {"seven": 0.5}
        }
        assert initial.trust == {"s1": 0, "s2": 0, "s3": 0}

        assert first.trust == {"s1": 1, "s2": 2 / 3, "s3": 1 / 3}
        assert first.belief == {
            "x": {"one": 1},
            "y": {"nine": 3 / 5, "eight": 2 / 5},
            "z": {"seven": 4 / 5}
        }



from enum import Enum
import re

from bidict import bidict

from truthdiscovery.algorithm import (
    AverageLog,
    Investment,
    MajorityVoting,
    PooledInvestment,
    PriorBelief,
    Sums,
    TruthFinder
)
from truthdiscovery.utils import (
    ConvergenceIterator,
    DistanceMeasures,
    filter_dict,
    FixedIterator
)


class OutputFields(Enum):
    """
    Fields available to show in results to a user
    """
    ACCURACY = "accuracy"
    BELIEF = "belief"
    BELIEF_STATS = "belief_stats"
    ITERATIONS = "iterations"
    MOST_BELIEVED = "most_believed_values"
    TIME = "time"
    TRUST = "trust"
    TRUST_STATS = "trust_stats"


class BaseClient:
    """
    Base class to provide functionality common to any interface to the library
    """
    ALG_LABEL_MAPPING = bidict({
        "average_log": AverageLog,
        "investment": Investment,
        "pooled_investment": PooledInvestment,
        "sums": Sums,
        "truthfinder": TruthFinder,
        "voting": MajorityVoting
    })

    def algorithm_cls(self, alg_label):
        """
        Return the algorithm class corresponding to a string label

        :param alg_label:     label in ``ALG_LABEL_MAPPING``
        :return:              a :any:`BaseAlgorithm` sub-class instance
        :raises ValueError:   if label is invalid
        """
        try:
            return self.ALG_LABEL_MAPPING[alg_label]
        except KeyError:
            raise ValueError(
                "invalid algorithm label '{}'".format(alg_label)
            )

    def algorithm_parameter(self, param_string):
        """
        Parse a string representation of a parameter to construct an algorithm
        object with

        :return: a pair ``(param_name, value)``
        """
        try:
            param, value = map(str.strip, param_string.split("=", maxsplit=1))
        except ValueError:
            raise ValueError(
                "parameters must be in the form 'key=value'"
            )
        # Map param name to a callable to convert string to correct type
        type_mapping = {
            "iterator": self.get_iterator,
            "priors": PriorBelief
        }
        type_convertor = type_mapping.get(param, float)
        return (param, type_convertor(value))

    def get_iterator(self, it_string, max_limit=200):
        """
        Parse an :any:`Iterator` object from a string representation
        """
        fixed_regex = re.compile(r"fixed-(?P<limit>\d+)$")
        convergence_regex = re.compile(
            r"(?P<measure>[^-]+)-convergence-(?P<threshold>[^-]+)"
            r"(-limit-(?P<limit>\d+))?$"  # optional limit
        )
        fixed_match = fixed_regex.match(it_string)
        if fixed_match:
            limit = int(fixed_match.group("limit"))
            if limit > max_limit:
                raise ValueError(
                    "Cannot perform more than {} iterations".format(max_limit)
                )
            return FixedIterator(limit=limit)

        convergence_match = convergence_regex.match(it_string)
        if convergence_match:
            measure_str = convergence_match.group("measure")
            try:
                measure = DistanceMeasures(measure_str)
            except ValueError:
                raise ValueError(
                    "invalid distance measure '{}'".format(measure_str)
                )
            threshold = float(convergence_match.group("threshold"))
            limit = max_limit
            if convergence_match.group("limit") is not None:
                limit = int(convergence_match.group("limit"))
                if limit > max_limit:
                    raise ValueError(
                        "Upper iteration limit cannot exceed {}"
                        .format(max_limit)
                    )
            return ConvergenceIterator(measure, threshold, limit)

        raise ValueError(
            "invalid iterator specification '{}'".format(it_string)
        )

    def get_algorithm_object(self, alg_cls, param_dict):
        """
        Instantiate an algorithm object
        """
        return alg_cls(**param_dict)

    def get_algorithm_params(self, alg_cls, param_dict):
        """
        :return: ``(params, ignored)``, where ``params`` is a dict containing
                 only the keys that are valid parameters for the given
                 algorithm class, and ``ignored`` is a set of the invalid
                 parameter names
        """
        filtered = dict(filter_dict(param_dict, alg_cls.get_parameter_names()))
        ignored = set(param_dict) - set(filtered)
        return filtered, ignored

    def get_ignored_parameters_message(self, cls, names):
        """
        Return a message indicating that invalid parameters have been ignored
        when instantiating an algorithm

        :param cls:   an :any:`BaseAlgorithm` sub-class
        :param names: iterable of parameter names
        :return: a message as a string
        """
        return("Ignored parameters for '{}': {}"
               .format(cls.__name__, ", ".join(sorted(names))))

    def get_output_obj(self, results, output_fields=None, sup_data=None):
        """
        Format a :any:`Result` class as a dictionary to present as output to
        the user. If ``output_fields`` is None, include all available fields
        """
        output_fields = output_fields or list(OutputFields)
        out = {}

        for field in output_fields:
            if field == OutputFields.TIME:
                out[field.value] = results.time_taken

            if field == OutputFields.ITERATIONS:
                out[field.value] = results.iterations

            if field == OutputFields.TRUST:
                out[field.value] = results.trust

            if field == OutputFields.BELIEF:
                out[field.value] = results.belief

            if field == OutputFields.TRUST_STATS:
                mean, stddev = results.get_trust_stats()
                out[field.value] = {"mean": mean, "stddev": stddev}

            if field == OutputFields.BELIEF_STATS:
                belief_stats = results.get_belief_stats()
                out[field.value] = {
                    var: {"mean": mean, "stddev": stddev}
                    for var, (mean, stddev) in belief_stats.items()
                }

            if sup_data is not None and field == OutputFields.ACCURACY:
                try:
                    acc = sup_data.get_accuracy(results)
                except ValueError:
                    acc = None
                out[field.value] = acc

            if field == OutputFields.MOST_BELIEVED:
                most_bel = {}
                for var in results.belief:
                    most_bel[var] = sorted(
                        results.get_most_believed_values(var)
                    )
                out[field.value] = most_bel

        return out



def _is_valid_file_type(data, signature):
    """
    :param data: file-like object
    :param signature: expected signature
    """
    size = len(signature)
    return data.read(size) == signature


def is_valid_png(data):
    return _is_valid_file_type(data, b"\x89\x50\x4E\x47\x0D\x0A\x1A\x0A")


def is_valid_gif(data):
    return _is_valid_file_type(data, b"GIF89a")



from truthdiscovery.input.dataset import Dataset
from truthdiscovery.input.supervised_data import SupervisedData


class FileDataset(Dataset):
    """
    Abstract class for loading datasets from files of a custom user-defined
    format
    """
    def __init__(self, filepath, *args, **kwargs):
        """
        :param filepath: path to file on disk to load dataset from
        """
        with open(filepath) as fileobj:
            super().__init__(self.get_tuples(fileobj), *args, **kwargs)

    def get_tuples(self, fileobj):
        """
        Parse a file and yield data tuples as required for the :any:`Dataset`
        constructor. This method must be implemented in child classes.

        :param fileobj: file object to read data from
        :yield: data tuples of the form ``(source_label, var_label, value)``
        """
        raise NotImplementedError("Must be implemented in child classes")


class FileSupervisedData(SupervisedData):
    """
    Abstract class for loading supervised data from files
    """
    def __init__(self, dataset, truth_path):
        """
        :param dataset: a :any:`Dataset` (or sub-class) object
        :param filepath: path to file on disk to load true values from
        """
        values = {}
        with open(truth_path) as truth_file:
            values = dict(self.get_pairs(truth_file))
        super().__init__(dataset, values)

    def get_pairs(self, fileobj):
        """
        Parse a file and yield the variables and true values. This method must
        be implemented in child classes.

        :param fileobj: file object to read from
        :yield: tuples of the form ``(var, true_value)``
        """
        raise NotImplementedError("Must be implemented in child classes")



import numpy as np

from truthdiscovery.algorithm.base import BaseIterativeAlgorithm
from truthdiscovery.exceptions import EarlyFinishError
from truthdiscovery.utils.iterator import ConvergenceIterator, DistanceMeasures


class TruthFinder(BaseIterativeAlgorithm):
    """
    TruthFinder by Yin, Han, Yu.
    """
    influence_param = 0.5
    dampening_factor = 0.3
    initial_trust = 0.9

    def __init__(self, *args, influence_param=None, dampening_factor=None,
                 initial_trust=None, **kwargs):
        """
        :param influence_param:  A number in [0, 1] that controls how much
                                 influence related claims have on confidence
                                 scores (rho in the paper)
        :param dampening_factor: A number in (0, 1) that prevents overly high
                                 confidence when sources are not independent
                                 (gamma in the paper)
        :param initial_trust:    Initial trust value for each source
        """
        if influence_param is not None:
            self.influence_param = influence_param
        if dampening_factor is not None:
            self.dampening_factor = dampening_factor
        if initial_trust is not None:
            self.initial_trust = initial_trust

        super().__init__(*args, **kwargs)

    def get_default_iterator(self):
        """
        Use cosine convergence iterator by default, as described in the paper
        """
        return ConvergenceIterator(DistanceMeasures.COSINE, 0.001)

    @classmethod
    def get_log_trust(cls, trust):
        """
        Return the 'tau' vector as defined in the TruthFinder paper. This
        involves taking logs to convert trust in [0, 1] to [0, +inf) to prevent
        numerical underflow

        :param trust: numpy array of trust values
        :return:      tau vector
        """
        if np.any(trust == 1):
            raise EarlyFinishError(
                "Trust has become 1 for at least one source"
            )
        return -np.log(1 - trust)

    def _run(self, data):
        trust = np.zeros((data.num_sources,))

        claim_counts = data.sc @ np.ones((data.num_claims),)
        # As in Investment, use multiply() to make sure the result is sparse
        a_mat = data.sc.T.multiply(1 / claim_counts).T
        b_mat = data.sc.T + self.influence_param * (data.imp.T @ data.sc.T)

        trust = np.full((data.num_sources,), self.initial_trust)
        belief = np.zeros((data.num_claims,))
        self.log(data, trust, belief)

        while not self.iterator.finished():
            try:
                log_belief = b_mat @ self.get_log_trust(trust)
            except EarlyFinishError:
                break
            belief = 1 / (1 + np.exp(-self.dampening_factor * log_belief))
            new_trust = a_mat @ belief
            self.iterator.compare(new_trust, trust)
            trust = new_trust
            self.log(data, trust, belief)

        return trust, belief



"""
Example script to produce a PNG image for a graph representation of a dataset
"""
import sys

from truthdiscovery.algorithm import Sums
from truthdiscovery.input import Dataset
from truthdiscovery.graphs import (
    GraphRenderer,
    PlainColourScheme,
    ResultsGradientColourScheme
)


if __name__ == "__main__":
    plain = False
    if len(sys.argv) == 3 and sys.argv[1] == "--plain":
        outpath = sys.argv[2]
        plain = True
    elif len(sys.argv) == 2:
        outpath = sys.argv[1]
    else:
        print("usage: {} [--plain] DEST".format(sys.argv[0]), file=sys.stderr)
        sys.exit(1)

    tuples = [
        ("source 1", "x", 4),
        ("source 1", "y", 7),
        ("source 2", "y", 7),
        ("source 2", "z", 5),
        ("source 3", "x", 3),
        ("source 3", "z", 5),
        ("source 4", "x", 3),
        ("source 4", "y", 6),
        ("source 4", "z", 8),
        ("my really long source name", "mylongvar", "extremelylongvalue"),
    ]
    mydata = Dataset(tuples)
    results = Sums().run(mydata)

    colour_scheme = (
        PlainColourScheme() if plain else ResultsGradientColourScheme(results)
    )
    renderer = GraphRenderer(
        width=1000, colours=colour_scheme
    )
    with open(outpath, "wb") as imgfile:
        renderer.render(mydata, imgfile)



from unittest.mock import Mock, patch

import numpy as np
import pytest

from truthdiscovery.algorithm import MajorityVoting, Sums
from truthdiscovery.input import Dataset
from truthdiscovery.output import Result, ResultDiff
from truthdiscovery.utils import FixedIterator


# Make a mocked time.time() function that returns increasing multiples of 5
MockedTime = Mock(side_effect=range(1, 100000, 5))


class TestResult:
    @pytest.fixture
    def example_results(self):
        trust = {
            "s1": 0.5,
            "s2": 0.7,
            "s3": 0.1
        }
        belief = {
            "x": {"red": 0.4, "blue": 0.9, "green": 0.5},
            "y": {"red": 0.1, "blue": 0.8, "green": 0},
            "z": {"red": 0.7, "blue": 0.7, "green": 1},
        }
        return Result(trust, belief, time_taken=0.5, iterations=100)

    def test_most_believed_values(self):
        test_data = (
            ({10: 0.5, 11: 0.7, 12: 0.7}, {11, 12}),
            ({-1: 0.1, -2: 0.1, 3: 0.05}, {-1, -2}),
            ({5: 0.9, 5000: 0.8}, {5}),
            ({123.456: 0.0001}, {123.456})
        )
        trust = [0.5] * 10
        for var_belief, exp in test_data:
            res = Result(trust, [var_belief], time_taken=None)
            assert set(res.get_most_believed_values(0)) == exp

    def test_num_iterations(self):
        data = Dataset([("source 1", "x", 7), ("source 2", "x", 8)])
        voting_res = MajorityVoting().run(data)
        assert voting_res.iterations is None

        sums_res = Sums(iterator=FixedIterator(13)).run(data)
        assert sums_res.iterations == 13

    @patch("time.time", MockedTime)
    def test_time_taken(self):
        """
        Test run time in Result objects for iterative and non-iterative
        algorithms
        """
        data = Dataset([("source 1", "x", 7), ("source 2", "x", 8)])
        res = MajorityVoting().run(data)
        assert res.time_taken == 5
        res = Sums().run(data)
        assert res.time_taken == 5

    def test_filter_result(self, example_results):
        res = example_results

        # Filter by sources
        source_filtered = res.filter(sources=("s1", "s3"))
        assert set(source_filtered.trust.keys()) == {"s1", "s3"}
        assert source_filtered.iterations == res.iterations
        assert source_filtered.time_taken == res.time_taken
        # belief should not be affected when only filtering on sources
        assert source_filtered.belief == res.belief
        # trust/belief dicts should be copies
        source_filtered.belief["x"]["red"] = 1000
        assert res.belief["x"]["red"] == 0.4

        # Filter by variables
        vars_filtered = res.filter(variables=("y", "z"))
        assert set(vars_filtered.belief.keys()) == {"y", "z"}
        assert vars_filtered.iterations == res.iterations
        assert vars_filtered.time_taken == res.time_taken
        assert vars_filtered.trust == res.trust

        # Filter both at the same time
        both_filtered = res.filter(sources=("s3", "s2"), variables=("z"))
        assert set(both_filtered.trust.keys()) == {"s2", "s3"}
        assert set(both_filtered.belief.keys()) == {"z"}

        # Shouldn't get an exception if a source/variable doesn't exist
        invalid_source = res.filter(sources=("joe", "s1"))
        invalid_var = res.filter(variables=("w", "x"))
        assert set(invalid_source.trust.keys()) == {"s1"}
        assert set(invalid_var.belief.keys()) == {"x"}

        # Edge case: filter set is empty
        empty_sources = res.filter(sources=[])
        empty_vars = res.filter(variables=[])
        assert not empty_sources.trust
        assert not empty_vars.belief
        assert empty_sources.belief == res.belief
        assert empty_vars.trust == res.trust

        # Edge case: no filtering
        no_filter = res.filter()
        assert no_filter.trust == res.trust
        assert no_filter.belief == res.belief

        # Should be able to use any iterable for filter set
        def mygen():
            yield "s1"
            yield "s3"
        gen_filter = res.filter(sources=mygen())
        assert set(gen_filter.trust.keys()) == {"s1", "s3"}

    def test_stats(self, example_results):
        res = example_results
        mean_trust, stddev_trust = res.get_trust_stats()
        assert mean_trust == (1.3 / 3)
        assert stddev_trust == 0.2494438257849294

        belief_stats = res.get_belief_stats()
        exp_x = (0.6, 0.21602468994692867)
        exp_y = (0.3, 0.3559026084010437)
        exp_z = (0.8, 0.14142135623730953)

        assert set(belief_stats.keys()) == {"x", "y", "z"}
        assert np.isclose(belief_stats["x"], exp_x).all()
        assert np.isclose(belief_stats["y"], exp_y).all()
        assert np.isclose(belief_stats["z"], exp_z).all()


class TestResultDiff:
    def test_no_common_sources_or_vars(self):
        res1 = Result(
            trust={"a": 0.1, "b": 0.2},
            belief={"x": {1: 1, 2: 0.5}, "y": {19: 0.4}},
            time_taken=125,
            iterations=10
        )
        res2 = Result(
            trust={"s": 0.1, "t": 0.2},
            belief={"u": {1: 1, 2: 0.5}, "v": {19: 0.4}},
            time_taken=75,
            iterations=14
        )
        diff = ResultDiff(res1, res2)
        assert diff.trust == {}
        assert diff.belief == {}
        assert diff.time_taken == -50
        assert diff.iterations == 4

    def test_common_sources(self):
        res1 = Result(
            trust={"a": 0.5, "b": 0.2, "c": 0.16},
            belief={"x": {1: 1, 2: 0.5}, "y": {19: 0.4}},
            time_taken=0.5,
            iterations=10
        )
        res2 = Result(
            trust={"s": 0.1, "t": 0.2, "a": 0.9, "c": 0.06},
            belief={"u": {1: 1, 2: 0.5}, "v": {19: 0.4}},
            time_taken=0.6,
            iterations=8
        )
        diff = ResultDiff(res1, res2)
        assert diff.trust == {"a": 0.4, "c": -0.1}
        assert diff.belief == {}

    def test_common_vars_but_no_common_values(self):
        res1 = Result(
            trust={"a": 0.1, "b": 0.2},
            belief={"x": {1: 1, 2: 0.5}, "y": {19: 0.4}},
            time_taken=0.5,
            iterations=10
        )
        res2 = Result(
            trust={"s": 0.1, "t": 0.2},
            belief={"u": {1: 1, 2: 0.5}, "v": {19: 0.4}, "x": {3: 1, 4: 0.5}},
            time_taken=0.6,
            iterations=8
        )
        diff = ResultDiff(res1, res2)
        assert diff.trust == {}
        assert diff.belief == {}

    def test_common_var_values(self):
        res1 = Result(
            trust={"a": 0.5, "b": 0.2, "c": 0.16},
            belief={"x": {1: 1, 2: 0.5}, "y": {19: 0.34}},
            time_taken=0.5,
            iterations=10
        )
        res2 = Result(
            trust={"s": 0.1, "t": 0.2, "a": 0.9, "c": 0.06},
            belief={"x": {1: 0.75, 2: 0.5}, "v": {19: 0.4}, "y": {19: 0.4}},
            time_taken=0.6,
            iterations=8
        )
        diff = ResultDiff(res1, res2)
        assert diff.trust == {"a": 0.4, "c": -0.1}
        assert diff.belief == {"x": {1: -0.25, 2: 0}, "y": {19: 0.06}}

    def test_no_iteration_info(self):
        res1 = Result(
            trust={"a": 0.1},
            belief={"x": {1: 1, 2: 0.5}},
            time_taken=0.5,
            iterations=10
        )
        res2 = Result(
            trust={"a": 0.1},
            belief={"x": {1: 1, 2: 0.5}},
            time_taken=0.5,
            iterations=None
        )
        diff = ResultDiff(res1, res2)
        assert diff.iterations is None



from io import StringIO
import json

from flask import Flask, render_template, request, jsonify

from truthdiscovery.algorithm import BaseIterativeAlgorithm
from truthdiscovery.client.base import BaseClient
from truthdiscovery.exceptions import ConvergenceError, EmptyDatasetError
from truthdiscovery.input import MatrixDataset
from truthdiscovery.output import Result, ResultDiff
from truthdiscovery.graphs import (
    JsonAnimator,
    JsonBackend,
    MatrixDatasetGraphRenderer,
    ResultsGradientColourScheme
)
from truthdiscovery.utils import DistanceMeasures


class route:
    """
    Class to be used a decorator instead of ``app.route`` which allows methods
    to be decorated.

    ``app.route`` cannot be used for methods since it sees an *unbound*
    function, whereas we wish for it to register as a method bound to a
    particular instance of the class.

    As such we keep a record of the parameters passed to the decorator and the
    method names, and defer actually adding the routes in flask until an
    instance has been created and passed to ``add_routes``
    """
    # Note: this list is shared amongst all ``route`` instances
    routes = []

    def __init__(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs

    def __call__(self, func):
        route.routes.append((self.args, self.kwargs, func.__name__))
        return func

    @classmethod
    def add_routes(cls, flask_app, class_instance):
        for args, kwargs, func_name in route.routes:
            try:
                meth = getattr(class_instance, func_name)
            except AttributeError:
                continue
            flask_app.add_url_rule(
                *args,
                view_func=meth,
                **kwargs
            )


class WebClient(BaseClient):
    def get_param_dict(self, params_str):
        """
        Parse a multi-line string of parameters to a dictionary where keys are
        parameter names, and values are parameter values as their proper Python
        types
        """
        params = {}
        if params_str is not None:
            for line in params_str.split("\n"):
                key, value = self.algorithm_parameter(line)
                params[key] = value
        return params

    def get_results_object(self, res_str):
        """
        Construct a :any:`Result` object from a JSON string

        :raises ValueError: if string is not valid JSON or does not contain the
                            fields required to construct a :any:`Result` object
        """
        try:
            obj = json.loads(res_str)
        except json.decoder.JSONDecodeError as ex:
            raise ValueError("invalid JSON: {}".format(ex))

        try:
            return Result(
                trust=obj["trust"],
                belief=obj["belief"],
                time_taken=obj["time"],
                iterations=obj["iterations"]
            )
        except KeyError as ex:
            raise ValueError("required field {} missing".format(ex))

    def get_graph_renderer(self, colours=None):
        """
        :param colours: (optional) :any:`GraphColourScheme` (or sub-class)
                        instance
        :return: a :any:`GraphRenderer` instance for graphs and animations
        """
        return MatrixDatasetGraphRenderer(
            width=800, node_radius=40, zero_indexed=False, colours=colours,
            font_size=20, backend=JsonBackend()
        )

    @route("/")
    def home_page(self):
        # Map algorithm labels to display name
        labels = {label: cls.__name__
                  for label, cls in self.ALG_LABEL_MAPPING.items()}
        # Data that the JS client needs to construct forms etc
        static_data = {
            "algorithm_labels": labels,
            "distance_measures": [m.value for m in DistanceMeasures]
        }
        return render_template("index.html", data_json=json.dumps(static_data))

    @route("/run/", methods=["GET"])
    def run(self):
        """
        Run an algorithm on a user-supplied dataset. Required HTTP parameters:
        * 'algorithm'
        * 'matrix'

        Optional parameters:
        * 'parameters'
        * 'previous_results'

        Responses are JSON objects of the form
        ``{"ok": True, "data": ...}``
        or
        ``{"ok": False, "error": ...}``
        """
        alg_labels = request.args.getlist("algorithm")
        matrix_csv = request.args.get("matrix")

        if not alg_labels or not matrix_csv:
            err_msg = "'algorithm' and 'matrix' parameters are required"
            return jsonify(ok=False, error=err_msg), 400

        matrix_csv = matrix_csv.replace("_", "")
        params_str = request.args.get("parameters")
        try:
            all_params = self.get_param_dict(params_str)
            dataset = MatrixDataset.from_csv(StringIO(matrix_csv))
        except ValueError as ex:
            return jsonify(ok=False, error=str(ex)), 400

        messages = []
        all_output = {}
        for alg_label in alg_labels:
            try:
                alg_cls = self.algorithm_cls(alg_label)
                params, ignored = self.get_algorithm_params(
                    alg_cls, all_params
                )
                alg = self.get_algorithm_object(alg_cls, params)
            except ValueError as ex:
                return jsonify(ok=False, error=str(ex)), 400

            # Show a message for each of the ignored parameters
            if ignored:
                msg = self.get_ignored_parameters_message(alg_cls, ignored)
                messages.append(msg)

            try:
                results = alg.run(dataset)
            except ConvergenceError as ex:
                return jsonify(ok=False, error=str(ex)), 500
            except EmptyDatasetError as ex:
                return jsonify(ok=False, error=str(ex)), 400

            output = self.get_output_obj(results)

            # Construct a graph and/or animation
            output["imagery"] = {}
            cs = ResultsGradientColourScheme(results)
            renderer = self.get_graph_renderer(colours=cs)
            json_buffer = StringIO()
            renderer.render(dataset, json_buffer)
            output["imagery"]["graph"] = json_buffer.getvalue()
            # Note: can only produce animation for iterative algorithms
            if isinstance(alg, BaseIterativeAlgorithm):
                animator = JsonAnimator(renderer=self.get_graph_renderer())
                json_buffer = StringIO()
                # Note: empty data and convergence error would already have
                # been caught above, so no need to check here
                animator.animate(
                    json_buffer, alg, dataset, show_progress=False
                )
                output["imagery"]["animation"] = json_buffer.getvalue()

            # Include diff between previous results if available
            prev_results = request.args.get("previous_results")
            if prev_results is not None:
                try:
                    obj = self.get_results_object(prev_results)
                except ValueError as ex:
                    err_msg = "'previous_results' is invalid: {}".format(ex)
                    return jsonify(ok=False, error=err_msg), 400

                # Previous results have been converted to JSON, which may have
                # changed numeric keys to strings: to ensure results can be
                # compared, convert the current results to and from JSON
                current_results = self.get_results_object(json.dumps(output))
                diff = ResultDiff(obj, current_results)
                output["diff"] = {
                    "trust": diff.trust,
                    "belief": diff.belief,
                    "time_taken": diff.time_taken,
                    "iterations": diff.iterations
                }

            all_output[alg_label] = output

        return jsonify({"ok": True, "data": all_output, "messages": messages})


def get_flask_app():
    client = WebClient()
    app = Flask(__name__)
    route.add_routes(app, client)
    return app


def run_debug_server():  # pragma: no cover
    app.run(host="0.0.0.0", debug=True)


app = get_flask_app()


if __name__ == "__main__":  # pragma: no cover
    run_debug_server()



import pytest

import numpy as np

from truthdiscovery.utils import (
    ConvergenceIterator,
    DistanceMeasures,
    FixedIterator,
    Iterator
)
from truthdiscovery.exceptions import ConvergenceError


class TestBaseIterator:
    def test_run_base_fail(self):
        it = Iterator()
        with pytest.raises(NotImplementedError):
            it.finished()

    def test_it_count(self):
        it = Iterator()
        assert it.it_count == 0

    def test_reset(self):
        limit = 25
        it = FixedIterator(limit)
        # Run the iterator down
        while not it.finished():
            it.compare(1, 2)
        # Reset: should no longer be finished
        it.reset()
        assert not it.finished()
        # Check can run it down again
        it_count = 0
        while not it.finished():
            it_count += 1
            it.compare(1, 2)
        assert it_count == limit
        assert it.it_count == limit

        # Perform the same test for a convergence iterator
        current_distance = 1
        conv_it = ConvergenceIterator(DistanceMeasures.L1, 0.501)
        while not conv_it.finished():
            current_distance -= 0.02
            conv_it.compare(np.array([1]), np.array([1 + current_distance]))
        conv_it.reset()
        assert not conv_it.finished()
        it_count = 0
        current_distance = 1
        while not conv_it.finished():
            current_distance -= 0.02
            it_count += 1
            conv_it.compare(np.array([1]), np.array([1 + current_distance]))
        assert it_count == limit
        assert conv_it.it_count == limit


class TestFixedIterator:
    def test_invalid_limit(self):
        invalid = (-1, -2)
        for limit in invalid:
            with pytest.raises(ValueError):
                FixedIterator(limit)

    def test_finish_condition(self):
        limit = 24
        it = FixedIterator(limit)
        it_count = 0
        # Careful: if implementation is completely broken the following may be
        # an infinite loop
        while not it.finished():
            # Do some stuf...
            it_count += 1
            # Objects compared here should not matter
            it.compare("hello", {"blah": 999})
        assert it_count == limit


class TestConvergenceIterator:
    def test_basic(self):
        threshold = 0.51
        it_count = 0
        it = ConvergenceIterator(DistanceMeasures.L1, threshold)
        current_distance = 1
        while not it.finished():
            it_count += 1
            current_distance -= 0.02
            it.compare(np.array([1]), np.array([1 + current_distance]))
        assert it_count == 25

    def test_invalid_distance_measures(self):
        invalid = ("L1", 1, 2, None)
        obj1 = np.array([1, 0.5, 0.4])
        obj2 = np.array([0.9, 0.8, 0.7])
        for measure in invalid:
            with pytest.raises(ValueError):
                ConvergenceIterator.get_distance(measure, obj1, obj2)

    def test_did_not_converge(self):
        it = ConvergenceIterator(DistanceMeasures.L_INF, 0.5, limit=200)
        it_count = 0
        with pytest.raises(ConvergenceError):
            while not it.finished():
                it_count += 1
                it.compare(np.array([1]), np.array([2]))
        assert it_count == 200


class TestDistanceMeasures:
    def check(self, measure, obj1, obj2, exp_distance):
        got = ConvergenceIterator.get_distance(
            measure, np.array(obj1), np.array(obj2)
        )
        assert got == exp_distance

    def test_l1(self):
        self.check(DistanceMeasures.L1, [1, 2, 3, 4], [0, 3, -4, 1], 12)

    def test_l2(self):
        self.check(
            DistanceMeasures.L2,
            [1, 2, 3, 4],
            [0, 3, -4, 1],
            7.745966692414834
        )

    def test_l_inf(self):
        self.check(DistanceMeasures.L_INF, [1, 2, 3, 4], [0, 3, -4, 1], 7)

    def test_cosine(self):
        self.check(
            DistanceMeasures.COSINE,
            [0, 1, 1.5, 2],
            [1, 1, 0.3, 0.2],
            0.5292255080098709
        )
        # Cosine similarity is not defined when one of the vectors is 0...
        # our distance should be 1 in this case
        self.check(
            DistanceMeasures.COSINE,
            [0, 0, 0, 0],
            [1, 1, 0.3, 0.2],
            1
        )
        self.check(
            DistanceMeasures.COSINE,
            [1, 1, 0.3, 0.2],
            [0, 0, 0, 0],
            1
        )



from truthdiscovery.algorithm.average_log import AverageLog
from truthdiscovery.algorithm.base import (
    BaseAlgorithm,
    BaseIterativeAlgorithm,
    PriorBelief
)
from truthdiscovery.algorithm.investment import Investment
from truthdiscovery.algorithm.pooled_investment import PooledInvestment
from truthdiscovery.algorithm.sums import Sums
from truthdiscovery.algorithm.truth_finder import TruthFinder
from truthdiscovery.algorithm.voting import MajorityVoting



from truthdiscovery.algorithm import *
from truthdiscovery.exceptions import *
from truthdiscovery.graphs import *
from truthdiscovery.input import *
from truthdiscovery.output import *
from truthdiscovery.utils import *



import math

import cairo

from truthdiscovery.graphs.backends.base import BaseBackend
from truthdiscovery.graphs.entities import Rectangle, Circle, Line


class PngBackend(BaseBackend):
    """
    Draw an image as a PNG using Cairo
    """
    def draw_entities(self, entities, outfile, width, height):
        # Initialise Cairo
        surface = cairo.ImageSurface(
            cairo.FORMAT_ARGB32, width, height
        )
        ctx = cairo.Context(surface)

        for ent in entities:
            ctx.set_source_rgba(*ent.colour)
            if isinstance(ent, Rectangle):
                ctx.rectangle(ent.x, ent.y, ent.width, ent.height)
                ctx.fill()

            if isinstance(ent, Circle):
                ctx.arc(ent.x, ent.y, ent.radius, 0, 2 * math.pi)
                ctx.fill()
                if ent.label is not None:
                    self.draw_label(ctx, ent.label, 2 * ent.radius, width)

            elif isinstance(ent, Line):
                if ent.dashed:
                    ctx.set_dash([10])  # TODO: configure dash size somehow
                else:
                    ctx.set_dash([])
                ctx.set_line_width(ent.width)
                ctx.move_to(ent.x, ent.y)
                ctx.line_to(ent.end_x, ent.end_y)
                ctx.stroke()

        surface.write_to_png(outfile)

    def draw_label(self, ctx, label, max_width, image_width):
        ctx.set_font_size(label.size)
        ext = ctx.text_extents(label.text)
        # Make sure long source or var names do not run off the screen
        label_lhs = min(image_width - ext.width, label.x - ext.width / 2)
        label_lhs = max(0, label_lhs)

        # If label exceeds the node, draw a background-coloured box where text
        # will go, so that the label does not clash with background/node border
        if ext.width > max_width:
            r, g, b = label.overflow_background
            ctx.set_source_rgba(*label.overflow_background)
            # Try and have label an 'opposite' colour for label to avoid
            # clashing
            label.colour = (1 - r, 1 - g, 1 - b)

            padding = 5  # allow some space between box border and text
            ctx.rectangle(
                label_lhs - padding,
                label.y - ext.height / 2 - padding,
                ext.width + 2 * padding,
                ext.height + 2 * padding
            )
            ctx.fill()

        ctx.set_source_rgba(*label.colour)
        ctx.move_to(label_lhs, label.y - ext.y_bearing / 2)
        ctx.show_text(label.text)



"""
Adapted from https://github.com/pypa/sampleproject
"""
from os import path

from setuptools import setup, find_packages

HERE = path.abspath(path.dirname(__file__))

with open(path.join(HERE, "README.md"), encoding="utf-8") as f:
    LONG_DESCRIPTION = f.read()

with open(path.join(HERE, "requirements.txt"), encoding="utf-8") as f:
    REQUIREMENTS = f.read().split("\n")

setup(
    name="truthdiscovery",
    version="1.0.1",
    description=("Python3 library implementing a selection of truth discovery"
                 "algorithms"),
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    url="https://github.com/joesingo/truthdiscovery",
    author="Joe Singleton",
    author_email="joesingo@gmail.com",
    packages=find_packages(exclude=["contrib", "doc", "tests"]),
    python_requires=">=3.5",
    # Note: this includes test and code style packages and requirements, which
    # should really go in extras_require...
    install_requires=REQUIREMENTS,
    entry_points={
        "console_scripts": [
            "truthdiscovery=truthdiscovery.client.cli:main"
        ]
    }
)



"""
Script to measure run-time of various algorithms as the number of sources and
variable increases
"""
from collections import OrderedDict
import json
import sys

import numpy as np
import matplotlib.pyplot as plt

from truthdiscovery.input import MatrixDataset, SyntheticData
from truthdiscovery.algorithm import (
    AverageLog,
    Investment,
    MajorityVoting,
    PooledInvestment,
    Sums,
    TruthFinder
)

# Sizes for sources/variables to use in the experiments
DATA_SIZES = list(range(100, 2001, 200))
# The fixed size for the parameter that is not being varied
FIXED_SIZE = 500
# Parameters for synthetic data generation
CLAIM_PROBABILITY = 0.1
DOMAIN_SIZE = 4

ALGORITHMS = OrderedDict({
    "voting": MajorityVoting(),
    "sums": Sums(),
    "average.log": AverageLog(),
    "investment": Investment(),
    "Pooled Investment": PooledInvestment(),
    "TruthFinder": TruthFinder()
})


def generate_timings():
    print("generating large trust vector...", file=sys.stderr)
    max_size = max(DATA_SIZES)
    trust = np.random.uniform(size=(max_size,))

    print("generating large dataset...", file=sys.stderr)
    large_synth = SyntheticData(
        trust,
        num_variables=max_size,
        claim_probability=CLAIM_PROBABILITY,
        domain_size=DOMAIN_SIZE
    )
    sv = large_synth.data.sv

    vary_sources_shapes = [(n, FIXED_SIZE) for n in DATA_SIZES]
    vary_vars_shapes = [(FIXED_SIZE, n) for n in DATA_SIZES]

    results = {
        "fixed_size": FIXED_SIZE,
        "vary_num_sources": perform_test(sv, vary_sources_shapes),
        "vary_num_vars": perform_test(sv, vary_vars_shapes)
    }
    print(json.dumps(results))


def perform_test(master_sv, data_shapes):
    """
    Perform a number of timing tests

    :param master_sv:   source-variabels matrix to extract smaller datasets
                        from
    :param data_shapes: an iterable of (num_sources, num_variables) for the
                        sizes of data to test with
    :return: an OrderedDict {alg_label: list_of_timings, ...}
    """
    timings = OrderedDict()
    for num_sources, num_vars in data_shapes:
        print(
            "getting reduced dataset: {} sources, {} variables..."
            .format(num_sources, num_vars),
            file=sys.stderr
        )
        data = MatrixDataset(master_sv[0:num_sources, 0:num_vars])

        for alg_label, alg in ALGORITHMS.items():
            print("  running {}...".format(alg_label), end="", file=sys.stderr)
            res = alg.run(data)
            print(" {:.3f} seconds".format(res.time_taken), file=sys.stderr)

            if alg_label not in timings:
                timings[alg_label] = []
            timings[alg_label].append(res.time_taken)
    return timings


def plot_results(timings):
    fig, axes = plt.subplots(1, 2)
    vary_sources_ax, vary_vars_ax = axes

    # Set titles and axis labels
    vary_sources_ax.set_title(
        "Increasing number of sources\n(number of variables fixed at {})"
        .format(timings["fixed_size"])
    )
    vary_sources_ax.set_xlabel("Number of sources")
    vary_sources_ax.set_ylabel("Run time (seconds)")
    vary_vars_ax.set_title(
        "Increasing number of variables\n(number of sources fixed at {})"
        .format(timings["fixed_size"])
    )
    vary_vars_ax.set_xlabel("Number of variables")
    vary_vars_ax.set_ylabel("Run time (seconds)")

    configurations = [
        (vary_sources_ax, "vary_num_sources"),
        (vary_vars_ax, "vary_num_vars")
    ]
    for ax, test_label in configurations:
        for alg_label, results in timings[test_label].items():
            ax.plot(
                DATA_SIZES, results, "o-", label=alg_label.capitalize(),
                linewidth=3
            )
        ax.legend()
    fig.tight_layout()
    plt.show()


def usage():
    print("usage: {} (generate | plot RESULTS)".format(sys.argv[0]),
          file=sys.stderr)
    print("generate timing results as JSON and print to stdout, or plot"
          "results from a JSON file", file=sys.stderr)


def main():
    if len(sys.argv) < 2:
        usage()
        sys.exit(1)

    if sys.argv[1] == "generate":
        generate_timings()
    elif sys.argv[1] == "plot" and len(sys.argv) >= 3:
        with open(sys.argv[2]) as jsonfile:
            plot_results(json.load(jsonfile))
    else:
        usage()
        sys.exit(1)


if __name__ == "__main__":
    main()



from io import BytesIO, StringIO
import json
from operator import attrgetter

import imageio
import pytest

from truthdiscovery.algorithm import Sums
from truthdiscovery.input import Dataset, MatrixDataset
from truthdiscovery.output import Result
from truthdiscovery.graphs import (
    BaseAnimator,
    BaseBackend,
    Circle,
    GifAnimator,
    GraphColourScheme,
    GraphRenderer,
    JsonAnimator,
    JsonBackend,
    Label,
    Line,
    MatrixDatasetGraphRenderer,
    NodeType,
    PlainColourScheme,
    PngBackend,
    Rectangle,
    ResultsGradientColourScheme,
)
from truthdiscovery.utils import FixedIterator
from truthdiscovery.test.utils import is_valid_png, is_valid_gif


class ExampleColourScheme(GraphColourScheme):
    colours = {
        NodeType.SOURCE: (0.1, 0.1, 0.1),
        NodeType.CLAIM: (0.2, 0.2, 0.2),
        NodeType.VARIABLE: (0.3, 0.3, 0.3),
        "edge": (0.4, 0.4, 0.4),
        "label": (0.5, 0.5, 0.5),
        "border": (0.6, 0.6, 0.6),
        "background": (0.7, 0.7, 0.7)
    }

    def get_node_colour(self, node_type, _hints):
        cs = self.colours
        return cs[node_type], cs["label"], cs["border"]

    def get_background_colour(self):
        return self.colours["background"]

    def get_edge_colour(self):
        return self.colours["edge"]


class BaseTest:
    @pytest.fixture
    def dataset(self):
        return Dataset((
            ("s1", "x", 0),
            ("s1", "y", 7),
            ("s2", "x", 4),
            ("s3", "y", 7),
            ("s3", "z", 9),
            ("s4", "x", 4),
        ))


class TestRendering(BaseTest):
    def test_entity_positioning(self, dataset):
        cs = ExampleColourScheme()
        rend = GraphRenderer(width=100, node_radius=10, spacing=14, colours=cs)
        ents = list(rend.compile(dataset))

        assert len(ents) == (
            1              # background
            + 4 * 2        # sources
            + 6            # edges from sources to claims
            + 4 * 2        # claims
            + 4            # edges from claims to variables
            + 3 * 2        # variables
            + 2 * (6 + 4)  # arrow heads
        )

        colours = ExampleColourScheme.colours
        bgs = [e for e in ents if e.colour == colours["background"]]
        assert len(bgs) == 1
        assert isinstance(bgs[0], Rectangle)
        assert bgs[0].x == 0
        assert bgs[0].y == 0
        assert bgs[0].width == 100

        # The max number of verical nodes is 4, so height should be
        # 2*radius*4 + spacing*3
        assert bgs[0].height == 2 * 10 * 4 + 14 * 3

        # Should be 4 sources, 4 claims, and 3 variables
        source_ents = [e for e in ents if e.colour == colours[NodeType.SOURCE]]
        claim_ents = [e for e in ents if e.colour == colours[NodeType.CLAIM]]
        var_ents = [e for e in ents if e.colour == colours[NodeType.VARIABLE]]
        for e in source_ents:
            print(e.x, e.y, e.colour)
        assert len(source_ents) == 4
        assert len(claim_ents) == 4
        assert len(var_ents) == 3

        # Should be 4 + 4 + 3 border circles
        border_ents = [e for e in ents if e.colour == colours["border"]]
        assert len(border_ents) == 4 + 4 + 3

        # All nodes should be Circle objects
        for e in source_ents + claim_ents + var_ents + border_ents:
            assert isinstance(e, Circle)

        # There should be 6 edges from sources to claims, and 4 from claims to
        # variables. For each edge there should be two lines for the arrow
        # heads
        edge_ents = [e for e in ents if e.colour == colours["edge"]]
        assert len(edge_ents) == (6 + 4) * 3
        for e in edge_ents:
            assert isinstance(e, Line)

        # Sources should be aligned in x coordinates
        source_coords = [(ent.x, ent.y) for ent in source_ents]
        assert len(set(x for x, y in source_coords)) == 1
        # Check y positioning
        y_coords = sorted(y for x, y in source_coords)
        assert y_coords == [10, 44, 78, 112]

    def test_node_ordering(self):
        # Construct a dataset where claim tuples are not sorted by source,
        # variable or value
        dataset = Dataset((
            ("source 4", "z", 9),
            ("source 3", "x", 4),
            ("source 1", "x", 1),
            ("source 1", "y", 2),
            ("source 2", "x", 6),
            ("source 2", "z", 3)
        ))

        # Claims should be sorted first by their variable's label, and then by
        # their own label. Change the default claim label method so that we can
        # distinguish between the correct behaviour and when the claims are
        # incorrectly sorted only by their own labels.
        class ExampleGraphRenderer(GraphRenderer):
            def get_claim_label(rend_self, var_id, val_hash):
                return "{val} = {var}".format(
                    val=rend_self.dataset.val_hashes.inverse[val_hash],
                    var=rend_self.dataset.var_ids.inverse[var_id]
                )

        rend = ExampleGraphRenderer(colours=ExampleColourScheme())
        ents = list(rend.compile(dataset))

        c = ExampleColourScheme.colours
        source_ls = [e.label for e in ents if e.colour == c[NodeType.SOURCE]]
        var_ls = [e.label for e in ents if e.colour == c[NodeType.VARIABLE]]
        claim_ls = [e.label for e in ents if e.colour == c[NodeType.CLAIM]]

        source_ls.sort(key=attrgetter("y"))
        var_ls.sort(key=attrgetter("y"))
        claim_ls.sort(key=attrgetter("y"))

        sorted_sources = [l.text for l in source_ls]
        sorted_vars = [l.text for l in var_ls]
        sorted_claims = [l.text for l in claim_ls]

        assert sorted_sources == [
            "source 1", "source 2", "source 3", "source 4"
        ]
        assert sorted_vars == ["x", "y", "z"]
        assert sorted_claims == [
            "1 = x", "4 = x", "6 = x",
            "2 = y",
            "3 = z", "9 = z"
        ]

    def test_single_source(self):
        one_source_dataset = Dataset((
            ("source 1", "x", 1),
            ("source 1", "y", 1)
        ))
        one_var_dataset = Dataset((
            ("source 1", "x", 1),
            ("source 2", "x", 2)
        ))
        one_claim_dataset = Dataset((
            ("source 1", "x", 1),
            ("source 2", "x", 1)
        ))
        rend = GraphRenderer(
            node_radius=66, spacing=13, colours=ExampleColourScheme()
        )
        # Check the positioning of nodes for the one-source dataset
        ents = list(rend.compile(one_source_dataset))
        colours = ExampleColourScheme.colours
        sources = [e for e in ents if e.colour == colours[NodeType.SOURCE]]
        assert len(sources) == 1
        assert sources[0].y == (2 * 66 * 2 + 13) / 2
        # Check the other two don't raise any exceptions
        list(rend.compile(one_var_dataset))
        list(rend.compile(one_claim_dataset))

    def test_png_is_default(self, dataset, tmpdir):
        out = tmpdir.join("mygraph.png")
        rend = GraphRenderer(backend=None)
        assert isinstance(rend.backend, PngBackend)
        rend.render(dataset, out)
        with open(str(out), "rb") as f:
            assert is_valid_png(f)

    def test_long_labels(self):
        dataset = Dataset((
            ("a-source-with-an-extremely-long-name", "x", 1000000000000000000),
            ("source 2", "quite a complicated variable name", 100)
        ))
        ents = list(GraphRenderer().compile(dataset))
        assert len(ents) == (
            1 +      # background
            6 * 2 +  # 6 nodes, each represented by two circles
            4 * 3    # 4 edges and arrowheads
        )

    def test_matrix_renderer(self):
        buf = StringIO()
        buf.write(",5,7\n,,\n1,2,3")
        buf.seek(0)
        dataset = MatrixDataset.from_csv(buf)
        rend1 = MatrixDatasetGraphRenderer()
        rend2 = MatrixDatasetGraphRenderer(zero_indexed=False)

        rend1.render(dataset, BytesIO())
        rend2.render(dataset, BytesIO())

        assert rend1.get_source_label(0) == "s0"
        assert rend2.get_source_label(0) == "s1"
        assert rend1.get_var_label(0) == "v1"
        assert rend2.get_var_label(0) == "v2"

        # Note that source 1 (in 0-index terms) makes no claims: ID 1 should
        # therefore be source 2 (in 0-index terms)
        assert rend1.get_source_label(1) == "s2"
        assert rend2.get_source_label(1) == "s3"

        assert rend1.get_claim_label(0, 1) == "v1=7"
        assert rend2.get_claim_label(0, 1) == "v2=7"

    def test_image_size(self, dataset):
        buf = BytesIO()
        w = 142

        nr = 13
        sp = 15
        h = 2 * nr * 4 + sp * 3

        print(dataset.num_sources)
        print(dataset.num_claims)
        print(dataset.num_variables)

        rend = GraphRenderer(
            width=w, node_radius=nr, spacing=sp, backend=PngBackend()
        )
        rend.render(dataset, buf)
        buf.seek(0)
        img_data = imageio.imread(buf)
        got_w, got_h, _ = img_data.shape
        assert (got_h, got_w) == (w, h)

    def test_progress_bar(self, dataset):
        w = 200
        anim_colour = (0.34, 0.99, 0.34)

        class MyColourScheme(GraphColourScheme):
            def get_animation_progress_colour(self):
                return anim_colour

        rend = GraphRenderer(width=w, colours=MyColourScheme())
        ents = list(rend.compile(dataset, animation_progress=0.25))
        rects = [e for e in ents if e.colour == anim_colour]
        assert len(rects) == 1
        assert rects[0].x == 0
        assert rects[0].width == 50

        # Test without progress bar
        ents2 = list(rend.compile(dataset, animation_progress=None))
        rects2 = [e for e in ents2 if e.colour == anim_colour]
        assert not len(rects2)

    def test_dashed_lines(self, dataset):
        class DashedRenderer(GraphRenderer):
            def is_dashed(rself, edge_hints):
                assert edge_hints is not None
                return True
        ents = (DashedRenderer().compile(dataset))
        lines = [e for e in ents if isinstance(e, Line)]
        dashed = [line for line in lines if line.dashed]

        num_dashed = len(dashed)
        num_non_dashed = len(lines) - num_dashed

        # Edges between nodes should be dashed, but arrow heads shouldn't be
        assert num_dashed == 10
        assert num_non_dashed == 20


class TestBackends(BaseTest):
    def test_base_backend(self, dataset, tmpdir):
        out = tmpdir.join("mygraph.png")
        with pytest.raises(NotImplementedError):
            GraphRenderer(backend=BaseBackend()).render(dataset, out)

    def test_valid_png(self, dataset, tmpdir):
        out = tmpdir.join("mygraph.png")
        GraphRenderer(backend=PngBackend()).render(dataset, out)
        with open(str(out), "rb") as f:
            assert is_valid_png(f)

    def test_results_based_valid_png(self, dataset, tmpdir):
        cs = ResultsGradientColourScheme(Sums().run(dataset))
        out = tmpdir.join("mygraph.png")
        GraphRenderer(backend=PngBackend(), colours=cs).render(dataset, out)
        with open(str(out), "rb") as f:
            assert is_valid_png(f)

    def test_json_backend(self):
        w = 123
        h = 2 * 10 * 2 + 5
        rend = GraphRenderer(
            width=w, node_radius=10, spacing=5, backend=JsonBackend()
        )
        buf = StringIO()

        data = Dataset((
            ("s1", "x", 0),
            ("s2", "y", 1)
        ))

        rend.render(data, buf)
        buf.seek(0)
        obj = json.load(buf)
        assert isinstance(obj, dict)
        assert "width" in obj
        assert "height" in obj
        assert "entities" in obj
        assert obj["width"] == w
        assert obj["height"] == h

        ents = obj["entities"]
        assert isinstance(ents, list)
        # Background
        assert len([e for e in ents if e["type"] == "rectangle"]) == 1
        # Nodes and borders
        assert len([e for e in ents if e["type"] == "circle"]) == 6 * 2
        # Edges and arrow heads
        assert len([e for e in ents if e["type"] == "line"]) == 3 * 4

        assert ents[0] == {
            "type": "rectangle",
            "x": 0,
            "y": 0,
            "colour": list(rend.colours.get_background_colour()),
            "width": w,
            "height": h
        }

    def test_json_backend_entity_serialisation(self):
        rect = Rectangle(x=1, y=2, colour=(0.5, 0.6, 0.7), width=10, height=20)
        assert JsonBackend.entity_to_dict(rect) == {
            "type": "rectangle",
            "x": 1,
            "y": 2,
            "colour": (0.5, 0.6, 0.7),
            "width": 10,
            "height": 20
        }
        circ = Circle(x=1, y=2, colour=(0.5, 0.6, 0.7), radius=11)
        assert JsonBackend.entity_to_dict(circ) == {
            "type": "circle",
            "x": 1,
            "y": 2,
            "colour": (0.5, 0.6, 0.7),
            "radius": 11,
            "label": None
        }
        line = Line(
            x=1, y=2, colour=(0.5, 0.6, 0.7), end_x=10, end_y=100, width=202
        )
        assert JsonBackend.entity_to_dict(line) == {
            "type": "line",
            "x": 1,
            "y": 2,
            "colour": (0.5, 0.6, 0.7),
            "end_x": 10,
            "end_y": 100,
            "width": 202,
            "dashed": False
        }
        label = Label(
            x=1, y=2, colour=(0.1, 0.4, 0.9), text="hello", size=123,
            overflow_background=(0.9, 0.4, 0.1)
        )
        assert JsonBackend.entity_to_dict(label) == {
            "type": "label",
            "x": 1,
            "y": 2,
            "colour": (0.1, 0.4, 0.9),
            "text": "hello",
            "size": 123,
            "overflow_background": (0.9, 0.4, 0.1)
        }
        cirl_with_label = Circle(
            x=1, y=2, colour=(0.5, 0.6, 0.7), radius=11, label=label
        )
        assert JsonBackend.entity_to_dict(cirl_with_label) == {
            "type": "circle",
            "x": 1,
            "y": 2,
            "colour": (0.5, 0.6, 0.7),
            "radius": 11,
            "label": JsonBackend.entity_to_dict(label)
        }


class TestColourSchemes:
    def test_get_gradient_colour(self):
        class Mock(ResultsGradientColourScheme):
            COLOURS = [1, 2, 3, 4]

        colour_scheme = Mock(None)
        test_cases = [
            # (level, expected index)
            (0, 0),
            (0.2, 0),
            (0.24999, 0),
            (0.25, 1),
            (0.49999, 1),
            (0.5, 2),
            (0.6, 2),
            (0.74999, 2),
            (0.75, 3),
            (0.999, 3),
            (1, 3)
        ]
        for level, exp_index in test_cases:
            colour, _ = colour_scheme.get_graded_colours(level)
            assert colour == colour_scheme.COLOURS[exp_index]

    def test_plain_colour_scheme(self):
        cs = PlainColourScheme()
        source = cs.get_node_colour(NodeType.SOURCE, 0)
        var = cs.get_node_colour(NodeType.VARIABLE, 0)
        claim = cs.get_node_colour(NodeType.CLAIM, 0, 0)
        # All nodes types should have the same colours in the plain theme
        assert source == var == claim
        # Label and border should be the same colour
        assert source[1] == source[2]
        # Edges should be the same colour as node borders
        edge = cs.get_edge_colour()
        assert edge == source[2]
        # Background should be transparent
        background = cs.get_background_colour()
        assert background[3] == 0

    def test_results_based_colours(self):
        results = Result(
            trust={"s1": 0, "s2": 0.01, "s3": 0.6, "s4": 1},
            belief={"x": {1: 0, 2: 0.34, 3: 0.7, 4: 1}},
            time_taken=None
        )
        colour_scheme = ResultsGradientColourScheme(results)
        n_s1, l_s1, _ = colour_scheme.get_node_colour(NodeType.SOURCE, "s1")
        n_s2, l_s2, _ = colour_scheme.get_node_colour(NodeType.SOURCE, "s2")
        n_s3, l_s3, _ = colour_scheme.get_node_colour(NodeType.SOURCE, "s3")
        n_s4, l_s4, _ = colour_scheme.get_node_colour(NodeType.SOURCE, "s4")

        assert n_s1 == n_s2
        assert l_s1 == l_s2
        assert l_s1 != l_s3
        assert l_s3 == l_s4

        claim_type = NodeType.CLAIM
        n_c1, l_c1, _ = colour_scheme.get_node_colour(claim_type, ("x", 1))
        n_c2, l_c2, _ = colour_scheme.get_node_colour(claim_type, ("x", 2))
        n_c3, l_c3, _ = colour_scheme.get_node_colour(claim_type, ("x", 3))
        n_c4, l_c4, _ = colour_scheme.get_node_colour(claim_type, ("x", 4))

        assert l_c1 == l_c2
        assert l_c1 != l_c3
        assert l_c3 == l_c4


class TestAnimations(BaseTest):
    def test_base(self, dataset):
        class MyAnimator(BaseAnimator):
            supported_backends = (PngBackend,)

        with pytest.raises(NotImplementedError):
            MyAnimator().animate(BytesIO(), Sums(), dataset)

    def test_gif_animation(self, dataset):
        w, h = 123, 95
        renderer = GraphRenderer(width=w, node_radius=10, spacing=5)
        animator = GifAnimator(renderer=renderer)
        alg = Sums()
        buf = BytesIO()
        animator.animate(buf, alg, dataset)
        buf.seek(0)
        assert is_valid_gif(buf)

        # Check dimensions are as expected
        buf.seek(0)
        img_data = imageio.imread(buf)
        got_w, got_h, _ = img_data.shape
        assert (got_h, got_w) == (w, h)

    def test_json_animation(self, dataset):
        w, h = 123, 95
        renderer = GraphRenderer(
            width=w, node_radius=10, spacing=5, backend=JsonBackend()
        )
        animator = JsonAnimator(renderer=renderer, frame_duration=1 / 9)
        alg = Sums(iterator=FixedIterator(4))
        buf = StringIO()
        animator.animate(buf, alg, dataset)
        buf.seek(0)
        obj = json.load(buf)

        assert "fps" in obj
        assert obj["fps"] == 9
        assert "frames" in obj
        assert isinstance(obj["frames"], list)
        assert len(obj["frames"]) == 5
        assert isinstance(obj["frames"][0], dict)
        assert "width" in obj["frames"][0]
        assert "height" in obj["frames"][0]
        assert "entities" in obj["frames"][0]
        assert obj["frames"][0]["width"] == w
        assert obj["frames"][0]["height"] == h

    def test_renderer(self):
        # Custom renderer should be used if provided
        custom_renderer = GraphRenderer(font_size=10000)
        anim = GifAnimator(renderer=custom_renderer)
        assert anim.renderer == custom_renderer

        # Otherwise a default renderer
        anim2 = GifAnimator()
        assert isinstance(anim2.renderer, GraphRenderer)

    def test_fps(self):
        anim = GifAnimator(frame_duration=1 / 3)
        assert anim.fps == 3

    def test_invalid_backend(self):
        png_rend = GraphRenderer(backend=PngBackend())
        json_rend = GraphRenderer(backend=JsonBackend())
        with pytest.raises(TypeError):
            GifAnimator(renderer=json_rend)
        with pytest.raises(TypeError):
            JsonAnimator(renderer=png_rend)

    def test_default_backend(self):
        class SomeClass:
            pass

        class MyAnimator(BaseAnimator):
            supported_backends = (SomeClass, float, int)

        anim = MyAnimator()
        assert isinstance(anim.renderer.backend, SomeClass)

    def test_progress_bar(self, dataset):
        w = 200
        rend = GraphRenderer(width=w, backend=JsonBackend())
        anim = JsonAnimator(renderer=rend)

        buf = StringIO()
        it = FixedIterator(20)
        alg = Sums(iterator=it)
        anim.animate(buf, alg, dataset, show_progress=True)
        buf.seek(0)
        obj = json.load(buf)
        # Get the frame for the 5th iteration, which is 1 / 4 through
        frame = obj["frames"][5]
        rects = [
            e for e in frame["entities"]
            if e["type"] == "rectangle" and e["width"] != w
        ]
        assert len(rects) == 1
        assert rects[0]["x"] == 0
        assert rects[0]["width"] == w / 4

        # Test without progress
        buf2 = StringIO()
        anim.animate(buf2, alg, dataset, show_progress=False)
        buf2.seek(0)
        obj2 = json.load(buf2)
        frame2 = obj2["frames"][5]
        rects2 = [
            e for e in frame2["entities"]
            if e["type"] == "rectangle" and e["width"] != w
        ]
        assert not len(rects2)



"""
Compare the speed of convergence for iterative algorithms.

This script runs iterative algorithms on a large synthetic dataset, and records
the distance between old and new trust vectors at each iteration.

These distances are then plotted, so that the convergence (or otherwise) of
each algorithm can be compared.
"""
from io import StringIO
from os import path
import sys

import matplotlib.pyplot as plt

from truthdiscovery.algorithm import (
    AverageLog,
    Investment,
    PooledInvestment,
    Sums,
    TruthFinder
)
from truthdiscovery.input import SupervisedData
from truthdiscovery.utils import ConvergenceIterator, DistanceMeasures
from truthdiscovery.exceptions import ConvergenceError


DATA_CSV = path.join(path.dirname(__file__), "large_synthetic_data.csv")
ALGORITHMS = [Sums, AverageLog, Investment, PooledInvestment, TruthFinder]
MEASURE = DistanceMeasures.L2


def main(csv_file):
    """
    Perform the test
    """
    print("Loading data...")
    sup = SupervisedData.from_csv(csv_file)
    fig, ax = plt.subplots()
    fig.suptitle(
        "Convergence experiment\n"
        "(synthetic data with {d.num_sources} sources, {d.num_variables} "
        "variables)".format(d=sup.data)
    )
    ax.set_xlabel("Iteration number")
    ax.set_ylabel(r"$\ell_2$ distance between old and new trust (log scale)")

    # map algorithm names to list of distances over time
    distances = {}
    iterator = ConvergenceIterator(MEASURE, 0, limit=100, debug=True)
    for cls in ALGORITHMS:
        name = cls.__name__
        print("running {} using {} measure".format(name, MEASURE))
        alg = cls(iterator=iterator)
        stdout = StringIO()
        sys.stdout = stdout
        try:
            _res = alg.run(sup.data)
        except ConvergenceError:
            pass
        finally:
            sys.stdout = sys.__stdout__

        distances[name] = []
        for line in stdout.getvalue().split("\n"):
            if not line:
                continue
            _, dist = line.split(",")
            distances[name].append(float(dist))

    max_its = max(len(dists) for dists in distances.values())
    x = range(1, max_its + 1)

    for name, dists in distances.items():
        while len(dists) < max_its:
            dists.append(None)
        ax.semilogy(x, dists, label=name, linewidth=3)
    ax.legend()
    plt.show()


if __name__ == "__main__":
    with open(DATA_CSV) as csv_file:
        main(csv_file)



import time

import numpy as np

from truthdiscovery.algorithm.base import BaseAlgorithm
from truthdiscovery.output import Result


class MajorityVoting(BaseAlgorithm):
    """
    Baseline truth discovery method, where the belief that X takes value v is
    simply the number of sources who makes that assertion. Sources are
    considered to be equally trustworthy.
    """
    def run(self, data):
        """
        :param data: input data as a :any:`Dataset` object
        :return: results as a :any:`Result` object
        """
        super().run(data)
        start_time = time.time()
        claim_belief = data.sc.T @ np.ones((data.num_sources),)
        normalised_belief = claim_belief / np.max(claim_belief)
        end_time = time.time()
        return Result(
            trust=data.get_source_trust_dict([1] * data.num_sources),
            belief=data.get_belief_dict(normalised_belief),
            time_taken=end_time - start_time
        )



import math

import numpy as np

from truthdiscovery.graphs.backends import PngBackend
from truthdiscovery.graphs.colours import NodeType, GraphColourScheme
from truthdiscovery.graphs.entities import Rectangle, Circle, Line, Label


def _sort_keys_by_value(dct):
    """
    :param dct: a dictionary
    :return: a list of the keys in ``dict`` sorted by their corresponding value
    """
    return sorted(dct, key=lambda key: dct[key])


class GraphRenderer:
    """
    Create an image that shows a graph representation of a truth discovery
    dataset
    """
    def __init__(self, width=800, node_radius=50, spacing=20, line_width=3,
                 node_border_width=3, font_size=15, colours=None,
                 backend=None):
        """
        :param width:             width of the image in pixels
        :param node_radius:       node radius in pixels
        :param spacing:           minimum vertical spacing between nodes in
                                  pixels
        :param node_border_width: width in pixels of node border (use None or 0
                                  for no border)
        :param line_width:        with in pixels for edges between nodes
        :param font_size:         font size for node labels
        :param colours:           :any:`GraphColourScheme` (or sub-class)
                                  instance
        :param backend:           a :any:`BaseBackend` object (default is PNG)
        """
        self.width = width
        # Height is calculated based on the number of nodes, so is set later
        self.height = None

        self.node_radius = node_radius
        self.spacing = spacing
        self.node_border_width = node_border_width or 0
        self.line_width = line_width
        self.font_size = font_size
        self.colours = colours or GraphColourScheme()
        self.backend = backend or PngBackend()

        self.dataset = None

    def _get_y_coord(self, index, num_nodes):
        if num_nodes > 1:
            available_height = self.height - 2 * self.node_radius
            return (self.node_radius
                    + available_height * index / (num_nodes - 1))
        return self.height / 2

    def get_source_coords(self, index):
        y = self._get_y_coord(index, self.dataset.num_sources)
        return (self.node_radius, y)

    def get_var_coords(self, index, claim_y_coords):
        y = np.mean(claim_y_coords)
        return (self.width - self.node_radius, y)

    def get_claim_coords(self, index):
        y = self._get_y_coord(index, self.dataset.num_claims)
        return (self.width / 2, y)

    def get_source_label(self, source_id):
        return self.dataset.source_ids.inverse[source_id]

    def get_var_label(self, var_id):
        return self.dataset.var_ids.inverse[var_id]

    def get_claim_label(self, var_id, val_hash):
        return "{var} = {val}".format(
            var=self.dataset.var_ids.inverse[var_id],
            val=self.dataset.val_hashes.inverse[val_hash]
        )

    def get_height(self, dataset):
        max_vertical_nodes = max(
            dataset.num_sources,
            dataset.num_claims,
            dataset.num_variables
        )
        nodes_px = max_vertical_nodes * 2 * self.node_radius
        spacing_px = (max_vertical_nodes - 1) * self.spacing
        return int(nodes_px + spacing_px)

    def render(self, dataset, outfile, animation_progress=None):
        """
        :param dataset:            a :any:`Dataset` object
        :param outfile:            file object to write to
        :param animation_progress: percentage animation progress in [0, 1] to
                                   display how far through an animation we are
                                   (None if not animating)
        """
        self.height = self.get_height(dataset)
        entities = self.compile(dataset, animation_progress=animation_progress)
        self.backend.draw_entities(entities, outfile, self.width, self.height)

    def compile(self, dataset, animation_progress=None):
        """
        A generator of :any:`Entity` objects describing what to draw and in
        which order
        """
        if not self.height:
            # This allows `compile` to be called directly in tests...
            self.height = self.get_height(dataset)

        yield from self.compile_background()
        self.dataset = dataset

        # Get labels for nodes
        source_labels = {}
        var_labels = {}
        claim_labels = {}

        for s_id in self.dataset.source_ids.values():
            source_labels[s_id] = self.get_source_label(s_id)
        for var_id in self.dataset.var_ids.values():
            var_labels[var_id] = self.get_var_label(var_id)
        for ((var_id, val_hash), claim_id) in self.dataset.claim_ids.items():
            claim_labels[claim_id] = self.get_claim_label(var_id, val_hash)

        # Get y-coordinates for all nodes. We sort nodes of each type by their
        # label to determine the ordering
        source_coords = {}
        var_coords = {}
        claim_coords = {}
        for i, s_id in enumerate(_sort_keys_by_value(source_labels)):
            source_coords[s_id] = self.get_source_coords(i)
        # For claims, we sort first by variable label and then by claim label:
        # this ensures that claims for the same variable are all next to each
        # other
        claim_sort_keys = {}
        for ((var_id, _), claim_id) in self.dataset.claim_ids.items():
            sort_key = (var_labels[var_id], claim_labels[claim_id])
            claim_sort_keys[claim_id] = sort_key
        for i, claim_id in enumerate(_sort_keys_by_value(claim_sort_keys)):
            claim_coords[claim_id] = self.get_claim_coords(i)
        # For variables, average use y-coords of the associated claims
        for i, var_id in enumerate(_sort_keys_by_value(var_labels)):
            claim_ys = [
                claim_coords[c_id][1]
                for (v_id, _), c_id in self.dataset.claim_ids.items()
                if v_id == var_id
            ]
            var_coords[var_id] = self.get_var_coords(i, claim_ys)

        # Draw edges between sources and claims
        for s_id, claim_id in np.transpose(np.nonzero(self.dataset.sc)):
            _, val_hash = self.dataset.claim_ids.inverse[claim_id]
            hint = (
                self.dataset.source_ids.inverse[s_id],     # source label
                self.dataset.val_hashes.inverse[val_hash]  # value
            )
            yield from self.compile_edge(
                source_coords[s_id], claim_coords[claim_id], hint
            )

        # Draw claims, variables and sources
        for (var_id, val_hash), claim_id in self.dataset.claim_ids.items():
            # Draw edge from claim to variable
            hint = (
                self.dataset.val_hashes.inverse[val_hash],  # value
                self.dataset.var_ids.inverse[var_id]        # variable name
            )
            yield from self.compile_edge(
                claim_coords[claim_id], var_coords[var_id], hint
            )

            # Draw claim
            hint = (
                self.dataset.var_ids.inverse[var_id],      # variable name
                self.dataset.val_hashes.inverse[val_hash]  # value
            )
            yield from self.compile_node(
                NodeType.CLAIM, claim_labels[claim_id], claim_coords[claim_id],
                hint
            )

        # Variables
        for var, var_id in self.dataset.var_ids.items():
            coords = var_coords[var_id]
            label = var_labels[var_id]
            yield from self.compile_node(NodeType.VARIABLE, label, coords, var)

        # Sources
        for source, s_id in self.dataset.source_ids.items():
            coords = source_coords[s_id]
            label = source_labels[s_id]
            yield from self.compile_node(
                NodeType.SOURCE, label, coords, source
            )

        # Draw animation progress bar if required
        if animation_progress is not None:
            yield from self.compile_animation_progress_rect(animation_progress)

    def compile_background(self):
        """
        Yield a Rectangle for the background
        """
        yield Rectangle(
            x=0, y=0, width=self.width, height=self.height,
            colour=self.colours.get_background_colour()
        )

    def compile_node(self, node_type, label, coords, node_hints):
        """
        Return a generator of :any:`Entity` objects for drawing a node.

        :param node_type:  value from :any:`NodeType` enumeration
        :param label:      text to draw for this node
        :param coords:     a tuple (x, y)
        :param node_hints: data that identifies the node, which is forwarded to
                           colour scheme to get node colour
        """
        node_c, label_c, border_c = self.colours.get_node_colour(
            node_type, node_hints
        )

        # Draw circle for the node
        x, y = coords
        # First solid border colour...
        if self.node_border_width > 0:
            yield Circle(x=x, y=y, colour=border_c, radius=self.node_radius)
        # ...then the interior
        label = Label(
            x=x, y=y, colour=label_c,
            text=str(label),
            size=self.font_size,
            overflow_background=self.colours.get_background_colour()
        )
        yield Circle(
            x=x, y=y, colour=node_c,
            radius=self.node_radius - self.node_border_width,
            label=label
        )

    def compile_edge(self, start, end, edge_hints):
        start_x, start_y = start
        end_x, end_y = end
        colour = self.colours.get_edge_colour()
        yield Line(
            x=start_x, y=start_y, colour=colour, end_x=end_x, end_y=end_y,
            width=self.line_width, dashed=self.is_dashed(edge_hints)
        )
        # Yield lines for arrow head
        arrow_size = 10
        arrow_angle = math.pi / 7
        edge_inclination = math.atan2(end_y - start_y, end_x - start_x)
        # Calculate the point where edge crosses the node circle
        vec = np.array([-self.node_radius, 0])
        vec = self.rotation_matrix(edge_inclination) @ vec
        node_crossing = np.array(end) + vec

        for s in (-1, 1):
            # Here we consider the vector that starts at `node_crossing` and
            # takes us along the top/bottom arrow head. Start by imagining
            # arrow head is horizontal, then rotate to get the desired angle
            # wrt the edge
            displacement = np.array([-arrow_size, 0])
            angle = s * arrow_angle + edge_inclination
            displacement = self.rotation_matrix(angle) @ displacement
            arrow_end = node_crossing + displacement
            yield Line(
                x=node_crossing[0], y=node_crossing[1], end_x=arrow_end[0],
                end_y=arrow_end[1], width=self.line_width, colour=colour
            )

    def compile_animation_progress_rect(self, progress):
        """
        Return a :any:`Rectangle` for a bar across the bottom of the image to
        indicate progress through an animation

        :param progress: number in [0, 1] indicating width of the bar as a
                         fraction of total image width
        """
        colour = self.colours.get_animation_progress_colour()
        size = 5
        yield Rectangle(
            x=0,
            y=self.height - size,
            colour=colour,
            width=self.width * progress,
            height=size
        )

    def is_dashed(self, edge_hints):
        """
        Return a bool indicating whether a given edge should be dashed

        :param edge_hints: forwarded hints about an edge
        """
        return False

    @classmethod
    def rotation_matrix(cls, angle):
        return np.array([
            [np.cos(angle), -np.sin(angle)],
            [np.sin(angle), np.cos(angle)],
        ])


class MatrixDatasetGraphRenderer(GraphRenderer):
    """
    Modify node labels for matrix datasets, where source/var labels are just
    ints
    """
    def __init__(self, *args, zero_indexed=True, **kwargs):
        self.zero_indexed = zero_indexed
        super().__init__(*args, **kwargs)

    def format_label(self, label):
        if not self.zero_indexed:
            label += 1
        return str(label)

    def get_source_label(self, source_id):
        label = super().get_source_label(source_id)
        return "s{}".format(self.format_label(label))

    def get_var_label(self, var_id):
        label = super().get_var_label(var_id)
        return "v{}".format(self.format_label(label))

    def get_claim_label(self, var_id, val_hash):
        val = str(self.dataset.val_hashes.inverse[val_hash])
        if val.endswith(".0"):
            val = val[:-2]
        return "{}={}".format(self.get_var_label(var_id), val)
